/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf;

import org.w3c.rdf.Triple;
import org.w3c.rdf.Resource;
import org.w3c.rdf.RDFnode;

import java.util.Enumeration;

/**
 * This is a programming interface to an RDF model.
 * An RDF model is a directed labeled graph.
 * This interface represents a model as set of triples.
 * Every triple has a subject, a predicate and an object.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public interface Model {

  // General

  /**
   * Set a base URI for the message.
   * Affects creating of new resources and serialization syntax.
   * <code>null</code> is just alright, meaning that
   * the serializer will use always rdf:about and never rdf:ID
   */
  public void setURI(String uri);

  /**
   * Returns current base URI setting.
   */
  public String getURI();

  // Model access

  /**
   * Number of triples in the model
   *
   * @return  number of triples
   */
  public int size();

  /**
   * Enumerate triples
   */
  public Enumeration elements();

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   * @see     org.w3c.rdf.Triple
   */
  public boolean contains(Triple t);

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model. A literal is created out of the string parameter <code>object</code>.
   * This method is just a shortcut.
   *
   * @see     org.w3c.rdf.Resource
   */
  public void add(Resource subject, Resource predicate, String object);

  /**
   * Adds a new triple to the model. The object of the triple is an RDFnode.
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFnode
   */
  public void add(Resource subject, Resource predicate, RDFnode object);

  /**
   * Adds a new triple to the model.
   *
   * @see     org.w3c.rdf.Triple
   */
  public void add( Triple t );

  /**
   * Removes the triple from the model.
   *
   * @see     org.w3c.rdf.Triple
   */
  public void remove(Triple t);

  /**
   * General method to search for triples.
   * <code>null</code> input for any parameter will match anything.
   * <p>Example: <code>Model result = m.find( null, RDF.type, new Resource("http://...#MyClass") )</code>
   * <p>finds all instances of the class <code>MyClass</code>
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFnode
   */
  public Model find( Resource subject, Resource predicate, RDFnode object );

  /**
   * Search for a single triple.
   * <code>null</code> input for any parameter will match anything.
   *
   * @return  a single triple if such was found;<br>
   *          <code>null</code> if nothing was found;<br>
   *          otherwise a RuntimeException is thrown.
   */
  public Triple find1( Resource subject, Resource predicate, RDFnode object );

  /**
   * Clone the model.
   */
  public Model duplicate();

  /**
   * Creates empty model of the same Class
   */
  public Model create();

  /**
   * Set union with another model.
   * @return  <code>this</code>, i.e. the model itself
   */
  public Model union(Model m);

  /**
   * Set difference with another model.
   * @return  <code>this</code>, i.e. the model itself
   */
  public Model difference(Model m);

  /**
   * Creates a unique unnamed resource
   */
  //  public Resource noname();
}

