/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.javagen;

import org.w3c.rdf.*;
import org.ginf.*;
import org.ginf.helpers.*;
import org.ginf.schemas.*;
import org.ginf.delivery.*;
import org.xml.sax.*;

import java.util.*;
import java.io.*;

/**
 *
 */
public class Main {

  static final int CMD_VALIDATE_ONLY = 0;
  static final int CMD_CONST = 1;
  static final int CMD_CLIENT = 2;
  static final int CMD_WRAPPER = 4;

  RDFFactory factory;

  String namespace;
  String outputDirectory;
  String packageDirectory;

  int cmd = CMD_VALIDATE_ONLY; // by default validate only

  String packageName = "default";
  String wrapperClass = "Wrapper";
  String constClass = "NS";
  String clientClass = "Client";
  Model schema, complete;

  boolean validate = true;

  public Main(RDFFactory f) {
    factory = f;
  }
  
  /**
   * Schema as input parameter
   */
  void generatePackage(String url, Model m) throws Exception {

    // FIXME: not every namespace is hash-based...
    //    namespace = RDFUtil.getNS(url);
    namespace = url;
    //    if(!namespace.endsWith("#"))
    //      namespace = namespace + "#";

    
    // complete schema information available
    complete = factory.createSchemaModel( factory.registry().get() );
    schema = factory.createSchemaModel( m );

    if(validate) {
      System.out.println("Validating schema...");
      ((SchemaModel)schema).validate();
    }

    //    System.out.println("\nMODEL AS TRIPLES:");
    //    RDFUtil.printTriples(m, System.out);

    //      System.out.println("\nMODEL AS XML:");
    //      parser.serialize( m, new OutputStreamWriter(System.out));

    // find package
    Resource pckg = null;

    try {
      pckg = complete.find1( new Resource( namespace ), RDF.type, GEN.Package ).subject();
    } catch (Exception any) {}

    if(pckg == null && cmd != CMD_VALIDATE_ONLY) {
      System.err.println("ERROR: No package information found about " + namespace);
      return;
    }
    
    try {
      packageName = complete.find1( pckg, GEN.packageName, null).object().toString(); 
    } catch (Exception any) {}
    
    try {
      wrapperClass = complete.find1( pckg, GEN.wrapperClass, null).object().toString();
    } catch (Exception any) {}
    
    try {
      constClass = complete.find1( pckg, GEN.constClass, null).object().toString();
    } catch (Exception any) {}
    
    try {
      clientClass = complete.find1( pckg, GEN.clientClass, null).object().toString();
    } catch (Exception any) {}

    System.err.println("Processing package: " + packageName + " into " + outputDirectory);

    prepareOutputDir();

    if(cmd == CMD_VALIDATE_ONLY || (cmd & CMD_CONST) != 0) {
      System.err.println("Generating schema/constants: " + packageName + "." + constClass);
      OutputStream out = openFileStream( constClass );
      dumpConstClass( out, schema, namespace );
      closeFileStream( out );
    }

    if((cmd & CMD_CLIENT) != 0) {
      System.err.println("Generating client: " + packageName + "." + clientClass);
    }

    if((cmd & CMD_WRAPPER) != 0) {

      System.err.println("Generating wrapper: " + packageName + "." + wrapperClass);

      Model m_req = schema.find( null, RDFS.subClassOf, CoreComm.Request );
      for(Enumeration en_req = m_req.elements(); en_req.hasMoreElements(); ) {
	
	Resource req = ((Triple)en_req.nextElement()).subject();
	System.err.println("Request: " + req);
	
	processParams( req );
      }
    }
  }

  void processParams( Resource req ) {

    // find all properties with domain req
    Model m_prop = schema.find( null, RDFS.domain, req );
    for(Enumeration en_prop = m_prop.elements(); en_prop.hasMoreElements(); ) {

      Resource prop = (Resource)((Triple)en_prop.nextElement()).subject();
      System.err.println("Domain property: " + prop + ", type=" + getPropType(prop));
      
      //      processParams( req );
    }
  }

  Resource getPropType(Resource prop) {
    
    // find all properties with domain req
    Triple t = schema.find1( prop, GEN.range, null );
    if(t != null)
      return (Resource)t.object();
    else
      return null;
  }

  static final String NS_IMPORT = "import org.ginf.*;\n" +
  "import org.ginf.helpers.*;\n" +
  "import org.w3c.rdf.*;\n" +
  "import org.xml.sax.*;\n" +
  "import java.io.*;\n";

  static final String NS_COMMENT = "/**\n" +
  " * This class provides handy access to schema information.\n" +
  " * THIS FILE HAS BEEN GENERATED AUTOMATICALLY.\n" +
  " */\n";

  static final String NS_NSDEF = "  /** Namespace URI of this schema */";

  static final String NS_LOADSCHEMA = "  public boolean loadSchema(String uri, RDFFactory factory) {\n" +
  "    if(NS.equals(uri)) {\n" +
  "      InputSource source = new InputSource( new StringReader( _schemaData ) );\n" +
  "      source.setSystemId(NS);\n" +
  "      factory.registry().fetch(source);\n" +
  "      return true;\n" +
  "    } else\n" +
  "      return false;\n" +
  "  }\n";

/*
  static final String NS_LOADSCHEMA = "  public static void loadSchema(RDFFactory factory) throws Exception {\n" + 
  "    InputSource source = new InputSource( new StringReader( _schemaData ) );\n" + 
  "    source.setSystemId( NS );\n" +
  "    Model m = factory.createRDFMS().parse( source );\n" +
  "    factory.registry().add( NS, m );\n" +
  "  }\n";
*/

  void dumpConstClass(OutputStream out, Model m, String ns) throws Exception {
    
    PrintWriter p = new PrintWriter( new OutputStreamWriter(out) );

    p.println("package " + packageName + ";\n");
    p.println(NS_IMPORT);
    p.println(NS_COMMENT);
    
    p.println("public class " + constClass + " implements SchemaLoader {\n");

    // find all resources
    //    Hashtable resources = new Hashtable();

    Model m_r = m.find( null, RDF.type, RDFS.Resource );
    //    m_r.union( m.find( null, RDF.type, RDF.Property ) );

    /*    
    for(Enumeration en = m.elements(); en.hasMoreElements(); ) {

      Triple t_r = (Triple)en.nextElement();
      resources.put( t_r.subject(), t_r.subject() );
      if( t_r.object() instanceof Resource )
	resources.put( t_r.object(), t_r.object() );
    }
    */
    // write namespace declaration

    p.println(NS_NSDEF);
    p.println("  public static final String NS = \"" + ns + "\";\n");

    // write resource declarations
    for(Enumeration en = m_r.elements(); en.hasMoreElements(); ) {
      
      Resource r = ((Triple)en.nextElement()).subject();
      String res = r.toString();
      //      String res = en.nextElement().toString();
      //      System.err.println("RESOURCE: " + res + ", ns=" + ns);
      if(res.startsWith(ns)) {
	String name = RDFUtil.getName(res);

	// comment?
	Triple tComment = RDFUtil.get1( m.find( r, RDFS.comment, null ) );
	if(tComment != null) {
	  p.println("  /** " + tComment.object() + " */");
	}
	
	p.println("  public static final Resource " + name + " = new Resource(NS + \"#" + name + "\");\n");
      }
    }

    // write schema

    // remove comments
    Model ignore = m.find( null, RDFS.comment, null );
    //    System.err.println("Comments: " + comments);
    Model m_out = m.duplicate().difference( ignore );
    m_out.difference( m.find( null, RDFS.label, null ) );
    

    m_out = RDFUtil.removeNamespace( GEN.NS, m_out );

    p.print("\n  static final String _schemaData = \"");
    p.flush();
    Writer w = new OutputStreamWriter( new StringEncoder(out));
    factory.createRDFMS().serialize( m_out, w );
    w.flush();
    p = new PrintWriter( new OutputStreamWriter(out) );
    p.println("\";\n");

    // public method for schema loading 

    p.println(NS_LOADSCHEMA);

    p.println("}");

    //    p.println("MODEL: " + m_out.getURI());

    p.flush();

  }

  void closeFileStream(OutputStream out) throws IOException {
    
    if(out != System.out)
      out.close();
  }

  OutputStream openFileStream( String cls ) throws IOException {

    if(outputDirectory == null)
      return System.out;

    File f = new File( packageDirectory + File.separator + cls + ".java");
    System.err.println("WRITING " + f);
    // replace existing file
    new RandomAccessFile(f, "rw").close();
    return new BufferedOutputStream( new FileOutputStream( f ) );
  }

  void prepareOutputDir() throws IOException {
    
    if(outputDirectory == null)
      return;
    
    File dir = new File( outputDirectory );
    if(!dir.isDirectory())
      throw new IOException("Invalid output directory: " + outputDirectory);
    
    // follow package names
    for(StringTokenizer st = new StringTokenizer(packageName, "."); st.hasMoreTokens();) {
      dir = new File( dir, st.nextToken() );
      if(!dir.exists())
	dir.mkdir();
      else if (!dir.isDirectory())
	throw new IOException("Cannot create directory: " + dir);
    }
    packageDirectory = dir.getPath();
  }

  public static void main(String[] args) {
    
    RDFFactory factory = new RDFFactoryImpl();
    AliasLoader al = new AliasLoader();
    factory.registry().addLoader(al);
      
    Main generator = new Main( factory );

    // parse options
    int par = 0;
    while (par < args.length) {

      if(args[par].startsWith("-d")) {
	
	generator.outputDirectory = args[par+1];
	par += 2;

      } else if(args[par].startsWith("-wr")) {
	generator.cmd |= CMD_WRAPPER;
	par++;
	
      } else if(args[par].startsWith("-sc")) {
	generator.cmd |= CMD_CONST;
	par++;
	
      } else if(args[par].startsWith("-cl")) {
	generator.cmd |= CMD_CLIENT;
	par++;
	
      } else if(args[par].startsWith("-all")) {
	generator.cmd |= CMD_WRAPPER | CMD_CONST | CMD_CLIENT;
	par++;

      } else if(args[par].startsWith("-ali")) {
	al.setAlias( args[par+1], args[par+2] );
	par += 3;

      } else if(args[par].startsWith("-nova")) {
	generator.validate = false;
	par++;

      } else if(args[par].startsWith("-us")) {

	// fetch additional schema
	factory.registry().fetch(args[par+1]);
	par += 2;

      } else
	break;
	
    }

    if(args.length == 0 || par != args.length - 1) {
      System.err.println("Usage: java org.ginf.javagen.Main  [-d <output directory>] [-novalidation] [-schema] [-wrapper] [-client] [-all] [-alias <original URI> <alias URI>] [-use <additional schema>]* <schema URI>\n" +
			 "Currently only -schema is supported. If this option is not specified, validation only.\n" +
			 "If you omit -d, generated class will be printed to stdout.");
      System.exit(1);
    }

    try {

      String schema = args[par];
      factory.registry().fetch(schema);

      generator.generatePackage( schema, factory.registry().get(schema) );
      
    } catch (Exception any) {
      System.err.println("Could not generate package: " + any);
      any.printStackTrace(System.err);
    }
  }


}
